// SPDX-License-Identifier: MIT

pragma solidity ^0.6.12;
pragma experimental ABIEncoderV2;

// OZ Imports

import { AdminUpgradeableProxy } from "../common/AdminUpgradeableProxy.sol";

// Tornado imports

import { ITornadoInstance } from "tornado-anonymity-mining/contracts/interfaces/ITornadoInstance.sol";

// Local V2 imports

import { FeeOracleManager } from "../v2/FeeOracleManager.sol";

import { InstanceRegistry } from "../v2/InstanceRegistry.sol";

import { CurveFeeOracle, ICurvePriceOracle, CurveChainedOracles } from "../v2/CurveFeeOracle.sol";

import { UniswapV3FeeOracle } from "../v2/UniswapV3FeeOracle.sol";

import { TornadoRouter } from "../v2/TornadoRouter.sol";

/**
 * @title CRVUSDInstancesProposal
 * @author AlienTornadosaurusHex
 * @notice Proposal to add crvUSD instances in 100, 1000, 10000, 100000, 1000000 denominations.
 */
contract CRVUSDInstancesProposal {
    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ VARIABLES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    /* @dev The first Curve pool which is capable of being used as a safe oracle due to analytical EMA price */
    address public constant crvUSDUSDCStableswap2Pool = 0x4DEcE678ceceb27446b35C672dC7d61F30bAD69E;

    /* @dev The second Curve pool which is capable of being used as a safe oracle due to analytical EMA price */
    address public constant tricryptoUSDCPool = 0x7F86Bf177Dd4F3494b841a37e810A34dD56c829B;

    /* @dev The InstanceRegistry contract, for registering instances, proxy must be upgraded */
    InstanceRegistry public constant instanceRegistry =
        InstanceRegistry(0xB20c66C4DE72433F3cE747b58B86830c459CA911);

    /* @dev The FeeOracleManager contract, proxy must be upgraded */
    FeeOracleManager public constant feeOracleManager =
        FeeOracleManager(0x5f6c97C6AD7bdd0AE7E0Dd4ca33A4ED3fDabD4D7);

    /* @dev This is the Uniswap V3 Oracle which we will use for all of our traditional instances, but it will
    also help the Curve instances, the former must have been deployed with the address of this */
    UniswapV3FeeOracle public immutable v3FeeOracle;

    /* @dev This is the CurveFeeOracle contract which will be deployed beforehand and which will be able to
    use multiple Curve pools as oracles, at once. */
    CurveFeeOracle public immutable curveFeeOracle;

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ LOGIC ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    constructor(address _deployedCurveFeeOracleAddress, address _deployedUniswapV3FeeOracleAddress) public {
        curveFeeOracle = CurveFeeOracle(_deployedCurveFeeOracleAddress);
        v3FeeOracle = UniswapV3FeeOracle(_deployedUniswapV3FeeOracleAddress);
    }

    /**
     * @dev This function also executes further internal functions inlined below.
     */
    function executeProposal() external {
        // Load the deployed crvUSD instances

        ITornadoInstance cu100 = ITornadoInstance(0x913a73486Dc4AA3832A56d461542836C1eeB93be);
        ITornadoInstance cu1_000 = ITornadoInstance(0x5A6b3C829dB3e938C885000c6E93CF35E74876a4);
        ITornadoInstance cu10_000 = ITornadoInstance(0x49f173CDAB99a2C3800F1255393DF9B7a17B82Bb);
        ITornadoInstance cu100_000 = ITornadoInstance(0x4640Dffc9fD0B113B983e3A350b070a119CA143C);
        ITornadoInstance cu1_000_000 = ITornadoInstance(0xc4eA8Bd3Fd76f3c255395793B47F7c55aD59d991);

        // Ok, first add the Uniswap V3 Oracle to the contract

        curveFeeOracle.setUniswapV3FeeOracle(v3FeeOracle);

        // Then, add necessary oracles for the CRVUSD price, to the CurveFeeOracle

        _setCurveFeeChainedOracleForInstance(curveFeeOracle, cu100);
        _setCurveFeeChainedOracleForInstance(curveFeeOracle, cu1_000);
        _setCurveFeeChainedOracleForInstance(curveFeeOracle, cu10_000);
        _setCurveFeeChainedOracleForInstance(curveFeeOracle, cu100_000);
        _setCurveFeeChainedOracleForInstance(curveFeeOracle, cu1_000_000);

        // Then, add the instances to the InstanceRegistry

        instanceRegistry.addInstance(cu100);
        instanceRegistry.addInstance(cu1_000);
        instanceRegistry.addInstance(cu10_000);
        instanceRegistry.addInstance(cu100_000);
        instanceRegistry.addInstance(cu1_000_000);

        // Finally, set all necessary data in the FeeOracleManager

        feeOracleManager.setFeeOracle(address(cu100), address(curveFeeOracle));
        feeOracleManager.setFeeOracle(address(cu1_000), address(curveFeeOracle));
        feeOracleManager.setFeeOracle(address(cu10_000), address(curveFeeOracle));
        feeOracleManager.setFeeOracle(address(cu100_000), address(curveFeeOracle));
        feeOracleManager.setFeeOracle(address(cu1_000_000), address(curveFeeOracle));

        // The fee will be 0.3 % for those pools for which it makes sense

        feeOracleManager.setFeePercentForInstance(cu100, 0);
        feeOracleManager.setFeePercentForInstance(cu1_000, 0);
        feeOracleManager.setFeePercentForInstance(cu10_000, 30);
        feeOracleManager.setFeePercentForInstance(cu100_000, 30);
        feeOracleManager.setFeePercentForInstance(cu1_000_000, 30);
    }

    /**
     * @dev This function adds pools as fee oracles to the main curve fee oracle contract. The oracles are
     * chained, this means that multiple pools can be used as one oracle using the library which can be found
     * in the contract file. This means:
     *
     * (...pool contracts as oracles) ===(`price_oracle`)===> CurveFeeOracle ===(`getFee()`)===>
     * FeeOracleManager ===(`instanceFeeWithUpdate`)===> RelayerRegistry
     */
    function _setCurveFeeChainedOracleForInstance(CurveFeeOracle _feeOracle, ITornadoInstance _instance)
        internal
    {
        // Add the oracles which are the USDC/CRVUSD stableswap and tricryptoUSDC pools

        ICurvePriceOracle[] memory _oracles = new ICurvePriceOracle[](2);

        _oracles[0] = ICurvePriceOracle(crvUSDUSDCStableswap2Pool);
        _oracles[1] = ICurvePriceOracle(tricryptoUSDCPool);

        // Add the selectors, so which functions to call because tricrypto pools supports getting prices for
        // multiple tokens, denominated in the first token, which is usually a stable

        bytes4[] memory _selectors = new bytes4[](2);

        _selectors[0] = CurveChainedOracles.PRICE_ORACLE_SELECTOR;
        _selectors[1] = CurveChainedOracles.PRICE_ORACLE_UINT256_SELECTOR;

        // Specify for the second that the ether price must be retrieved

        uint8[] memory _coins = new uint8[](2);

        _coins[1] = 1; // ETHER

        // In order to receive the CRVUSD price, its price in USDC must be read out (the USDC price in crvUSD
        // is worthless because we do not have a common denominator in the oracle then) and then the second
        // price must be inverted to receive the accurate ETH/CRVUSD price, meaning ETH per CRVUSD, which will
        // then be divided by ETH per TORN, and then (ETH/CRVUSD)/(ETH/TORN) = (ETH/CRVUSD)*(TORN/ETH) =
        // (TORN/CRVUSD) which is the price that the oracle should be supplying

        bool[] memory _invert = new bool[](2);

        _invert[0] = false; // DO NOT INVERT, PRICE IS CRVUSD IN USDC
        _invert[1] = true; // INVERT, PRICE IS ETH IN USDC, BUT WE NEED USDC IN ETH

        // (USDC/CRVUSD)*(ETH/USDC) = (ETH/CRVUSD)

        _feeOracle.modifyChainedOracleForInstance(
            _instance, _oracles, _selectors, _coins, _invert, "ETH/CRVUSD"
        );
    }
}
