// SPDX-License-Identifier: MIT

pragma solidity ^0.6.12;
pragma experimental ABIEncoderV2;

// OZ Imports

import { IERC20 } from "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import { Initializable } from "@openzeppelin/contracts/proxy/Initializable.sol";

// Tornado Imports

import { ITornadoInstance } from "tornado-anonymity-mining/contracts/interfaces/ITornadoInstance.sol";

// Local Imports

import { InstanceRegistry } from "./InstanceRegistry.sol";

import { IFeeOracle } from "./interfaces/IFeeOracle.sol";

/**
 * @title FeeManagerLegacyStorage
 * @author AlienTornadosaurusHex
 * @dev This is contract will help us layout storage properly for a proxy upgrade for the impl
 * FeeOracleManager (formerly FeeManager).
 */
contract FeeManagerLegacyStorage {
    /* @dev From first contract */
    uint24 private _deprecatedUniswapTornPoolSwappingFee;

    /* @dev From first contract */
    uint32 private _deprecatedUniswapTimePeriod;

    /* @dev From first contract, the only value we keep alive */
    uint24 public feeUpdateInterval;

    /* @dev From first contract, only used for initialization to preserve old values */
    mapping(ITornadoInstance => uint160) internal _oldFeesForInstance;

    /* @dev From first contract, only used for initialization to preserve old values */
    mapping(ITornadoInstance => uint256) internal _oldFeesForInstanceUpdateTime;
}

/**
 * @notice Fee data which is valid and also influences fee updating across all oracles.
 */
struct FeeData {
    uint160 feeAmount;
    uint64 feePercent;
    uint32 lastUpdated;
}

/**
 * @title FeeOracleManager
 * @author AlienTornadosaurusHex
 * @notice A contract which manages fee oracles and received data for other contracts to consume.
 * @dev This is an improved version of the FeeManager with a modified design from the original contract.
 */
contract FeeOracleManager is FeeManagerLegacyStorage, Initializable {
    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ VARIABLES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    /* @dev Divide protocol fee by this to get the percent value */
    uint64 public constant FEE_PERCENT_DIVISOR = 1 ether;

    /* @dev The Governance Proxy address */
    address public immutable governanceProxyAddress;

    /* @dev The TORN token */
    IERC20 public immutable torn;

    /* @dev The InstanceRegistry contract */
    InstanceRegistry public instanceRegistry;

    /* @dev Each instance has a dedicated fee oracle, these only compute the values */
    mapping(ITornadoInstance => IFeeOracle) public instanceFeeOracles;

    /* @dev The data for each instance will be stored in this contract */
    mapping(ITornadoInstance => FeeData) public feeDataForInstance;

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ EVENTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    event FeeUpdateIntervalUpdated(uint24 newLimit);
    event FeeUpdated(address indexed instance, uint256 newFee);
    event OracleUpdated(address indexed instance, address oracle);
    event InstanceFeePercentUpdated(address indexed instance, uint64 newFeePercent);
    event InstanceRegistryUpdated(address newAddress);

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ LOGIC ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    constructor(address _tornTokenAddress, address _governanceProxyAddress) public {
        torn = IERC20(_tornTokenAddress);
        governanceProxyAddress = _governanceProxyAddress;
    }

    modifier onlyGovernance() {
        require(msg.sender == governanceProxyAddress, "FeeOracleManager: onlyGovernance");
        _;
    }

    /**
     * @dev If there will be a need to initialize the proxy again, simply pad storage and inherit again,
     * making sure to not reference old data anywhere.
     */
    function initialize(
        address _uniswapV3FeeOracle,
        address _instanceRegistryAddress,
        ITornadoInstance[] calldata _instances,
        uint256[] calldata _feePercents
    ) external onlyGovernance initializer {
        // Get num of existing instances
        uint256 numInstances = _instances.length;

        for (uint256 i = 0; i < numInstances; i++) {
            // For each instance
            ITornadoInstance instance = _instances[i];

            // Store it's old data and the percent fees which Governance will command
            feeDataForInstance[instance] = FeeData({
                feeAmount: _oldFeesForInstance[instance],
                feePercent: uint64(_feePercents[i]),
                lastUpdated: uint32(_oldFeesForInstanceUpdateTime[instance])
            });

            // All old pools use the uniswap fee oracle
            instanceFeeOracles[instance] = IFeeOracle(_uniswapV3FeeOracle);
        }

        // Finally also store the instance registry
        instanceRegistry = InstanceRegistry(_instanceRegistryAddress);
    }

    function instanceFeeWithUpdate(ITornadoInstance _instance) public virtual returns (uint160) {
        return updateFee(_instance, true);
    }

    function updateAllFees(bool _respectFeeUpdateInterval)
        public
        virtual
        returns (uint160[] memory newFees)
    {
        return updateFees(instanceRegistry.getAllInstances(), _respectFeeUpdateInterval);
    }

    function updateFees(ITornadoInstance[] memory _instances, bool _respectFeeUpdateInterval)
        public
        virtual
        returns (uint160[] memory newFees)
    {
        uint256 numInstances = _instances.length;

        newFees = new uint160[](numInstances);

        for (uint256 i = 0; i < numInstances; i++) {
            newFees[i] = updateFee(_instances[i], _respectFeeUpdateInterval);
        }
    }

    function updateFee(ITornadoInstance _instance, bool _respectFeeUpdateInterval)
        public
        virtual
        returns (uint160)
    {
        // First get fee data
        FeeData memory feeData = feeDataForInstance[_instance];

        // Now update if we do not respect the interval or we respect it and are in the interval
        if (!_respectFeeUpdateInterval || (feeUpdateInterval < -feeData.lastUpdated + now)) {
            IFeeOracle oracle = instanceFeeOracles[_instance];

            // Check whether the instance is registered
            require(address(oracle) != address(0), "FeeOracleManager: instance has no oracle");

            // There must a be a fee set otherwise it's just 0
            if (feeData.feePercent != 0) {
                feeData.feeAmount = oracle.getFee(
                    torn,
                    _instance,
                    instanceRegistry.getInstanceData(_instance),
                    feeData.feePercent,
                    FEE_PERCENT_DIVISOR
                );
            } else {
                feeData.feeAmount = 0;
            }

            // Store
            feeDataForInstance[_instance] = FeeData({
                feeAmount: feeData.feeAmount,
                feePercent: feeData.feePercent,
                lastUpdated: uint32(now)
            });

            // Log
            emit FeeUpdated(address(_instance), feeData.feeAmount);
        }

        // In any case return a value
        return feeData.feeAmount;
    }

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ SETTERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    function setInstanceRegistry(address _newInstanceRegistryProxyAddress) external onlyGovernance {
        instanceRegistry = InstanceRegistry(_newInstanceRegistryProxyAddress);
        emit InstanceRegistryUpdated(_newInstanceRegistryProxyAddress);
    }

    function setFeeOracle(address _instanceAddress, address _oracleAddress) external onlyGovernance {
        // Prepare all contracts
        ITornadoInstance instance = ITornadoInstance(_instanceAddress);
        IFeeOracle oracle = IFeeOracle(_oracleAddress);

        // Nominally fee percent should be set first for an instance, but we cannot be sure
        // whether fee percent 0 is intentional, so we don't check
        FeeData memory feeData = feeDataForInstance[instance];

        // Fee which may be recalculated
        uint160 fee;

        // An address(0) oracle means we're removing an oracle for an instance
        if (_oracleAddress != address(0)) {
            // Reverts if oracle does not conform to interface
            fee = oracle.getFee(
                torn,
                instance,
                instanceRegistry.getInstanceData(instance),
                feeData.feePercent,
                FEE_PERCENT_DIVISOR
            );

            // Note down updated fee
            feeDataForInstance[instance] =
                FeeData({ feeAmount: fee, feePercent: feeData.feePercent, lastUpdated: uint32(now) });
        }

        // Ok, set the oracle
        instanceFeeOracles[instance] = oracle;

        // Logs
        emit OracleUpdated(_instanceAddress, _oracleAddress);
        emit FeeUpdated(_instanceAddress, fee);
    }

    function setFeePercentForInstance(ITornadoInstance _instance, uint64 _newFeePercent)
        external
        onlyGovernance
    {
        feeDataForInstance[_instance].feePercent = _newFeePercent;
        emit InstanceFeePercentUpdated(address(_instance), _newFeePercent);
    }

    function setFeeUpdateInterval(uint24 newLimit) external onlyGovernance {
        feeUpdateInterval = newLimit;
        emit FeeUpdateIntervalUpdated(newLimit);
    }

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GETTERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    function version() public pure virtual returns (string memory) {
        return "v2-oracle-manager";
    }

    function getUpdatedFeeForInstance(ITornadoInstance instance) public view virtual returns (uint160) {
        return instanceFeeOracles[instance].getFee(
            torn,
            instance,
            instanceRegistry.getInstanceData(instance),
            feeDataForInstance[instance].feePercent,
            FEE_PERCENT_DIVISOR
        );
    }

    function getLastFeeForInstance(ITornadoInstance instance) public view virtual returns (uint160) {
        return feeDataForInstance[instance].feeAmount;
    }

    function getLastUpdatedTimeForInstance(ITornadoInstance instance) public view virtual returns (uint32) {
        return feeDataForInstance[instance].lastUpdated;
    }

    function getFeePercentForInstance(ITornadoInstance instance) public view virtual returns (uint64) {
        return feeDataForInstance[instance].feePercent;
    }

    function getAllFeeDeviations() public view virtual returns (int256[] memory) {
        return getFeeDeviationsForInstances(instanceRegistry.getAllInstances());
    }

    function getFeeDeviationsForInstances(ITornadoInstance[] memory _instances)
        public
        view
        virtual
        returns (int256[] memory deviations)
    {
        uint256 numInstances = _instances.length;

        deviations = new int256[](numInstances);

        for (uint256 i = 0; i < numInstances; i++) {
            ITornadoInstance instance = _instances[i];

            FeeData memory feeData = feeDataForInstance[instance];

            uint256 marketFee = instanceFeeOracles[instance].getFee(
                torn,
                instance,
                instanceRegistry.getInstanceData(instance),
                feeData.feePercent,
                FEE_PERCENT_DIVISOR
            );

            int256 deviation;

            if (marketFee != 0) {
                deviation = int256((feeData.feeAmount * 1000) / marketFee) - 1000;
            }

            deviations[i] = deviation;
        }
    }
}
