// SPDX-License-Identifier: MIT

pragma solidity ^0.6.12;
pragma experimental ABIEncoderV2;

// OZ Imports

import { IERC20 } from "@openzeppelin/contracts/token/ERC20/IERC20.sol";
import { Initializable } from "@openzeppelin/contracts/proxy/Initializable.sol";
import { SafeERC20 } from "@openzeppelin/contracts/token/ERC20/SafeERC20.sol";
import { Math } from "@openzeppelin/contracts/math/Math.sol";

// Tornado Imports

import { ITornadoInstance } from "tornado-anonymity-mining/contracts/interfaces/ITornadoInstance.sol";

// Local imports

import { RelayerRegistry } from "../v1/RelayerRegistry.sol";
import { InstanceRegistry, InstanceData } from "./InstanceRegistry.sol";

/**
 * @title TornadoRouter
 * @author AlienTornadosaurusHex
 * @notice This contracts is a router for all Tornado Cash deposits and withdrawals
 * @dev This is an improved version of the TornadoRouter with a modified design from the original contract.
 */
contract TornadoRouter is Initializable {
    using SafeERC20 for IERC20;

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ VARIABLES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    /* @dev The address of the Governance proxy */
    address public immutable governanceProxyAddress;

    /* @dev The instance registry */
    InstanceRegistry public instanceRegistry;

    /* @dev The relayer registry */
    RelayerRegistry public relayerRegistry;

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ EVENTS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    event EncryptedNote(address indexed sender, bytes encryptedNote);
    event TokenApproved(address indexed spender, uint256 amount);

    event InstanceRegistryUpdated(address newInstanceRegistryProxyAddress);
    event RelayerRegistryUpdated(address newRelayerRegistryProxyAddress);

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ LOGIC ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    constructor(address _governanceProxyAddress) public {
        governanceProxyAddress = _governanceProxyAddress;
    }

    modifier onlyInstanceRegistry() {
        require(msg.sender == address(instanceRegistry), "TornadoRouter: onlyInstanceRegistry");
        _;
    }

    modifier onlyGovernance() {
        require(msg.sender == governanceProxyAddress, "TornadoRouter: onlyGovernance");
        _;
    }

    function initialize(address _instanceRegistryAddress, address _relayerRegistryAddress)
        external
        onlyGovernance
        initializer
    {
        instanceRegistry = InstanceRegistry(_instanceRegistryAddress);
        relayerRegistry = RelayerRegistry(_relayerRegistryAddress);
    }

    function deposit(ITornadoInstance _tornado, bytes32 _commitment, bytes calldata _encryptedNote)
        public
        payable
        virtual
    {
        (IERC20 token,, bool isERC20, bool isEnabled) = instanceRegistry.instanceData(_tornado);

        require(isEnabled, "TornadoRouter: instance not enabled");

        if (isERC20) {
            token.safeTransferFrom(msg.sender, address(this), _tornado.denomination());
        }

        _tornado.deposit{ value: msg.value }(_commitment);

        emit EncryptedNote(msg.sender, _encryptedNote);
    }

    function withdraw(
        ITornadoInstance _tornado,
        bytes calldata _proof,
        bytes32 _root,
        bytes32 _nullifierHash,
        address payable _recipient,
        address payable _relayer,
        uint256 _fee,
        uint256 _refund
    ) public payable virtual {
        (,,, bool isEnabled) = instanceRegistry.instanceData(_tornado);

        require(isEnabled, "TornadoRouter: instance not enabled");

        relayerRegistry.burn(msg.sender, _relayer, _tornado);

        _tornado.withdraw{ value: msg.value }(
            _proof, _root, _nullifierHash, _recipient, _relayer, _fee, _refund
        );
    }

    function backupNotes(bytes[] calldata _encryptedNotes) public virtual {
        for (uint256 i = 0; i < _encryptedNotes.length; i++) {
            emit EncryptedNote(msg.sender, _encryptedNotes[i]);
        }
    }

    function rescueTokens(IERC20 _token, address payable _to, uint256 _amount)
        public
        virtual
        onlyGovernance
    {
        require(_to != address(0), "TORN: can not send to zero address");

        if (_token == IERC20(0)) {
            // For Ether
            uint256 totalBalance = address(this).balance;
            uint256 balance = Math.min(totalBalance, _amount);
            _to.transfer(balance);
        } else {
            // For any other ERC20
            uint256 totalBalance = _token.balanceOf(address(this));
            uint256 balance = Math.min(totalBalance, _amount);
            require(balance > 0, "TORN: trying to send 0 balance");
            _token.safeTransfer(_to, balance);
        }
    }

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ SETTERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    function approveTokenForInstance(IERC20 _token, address _spender, uint256 _amount)
        external
        onlyInstanceRegistry
    {
        _token.safeApprove(_spender, _amount);
        emit TokenApproved(_spender, _amount);
    }

    function setInstanceRegistry(address _newInstanceRegistryProxyAddress) external onlyGovernance {
        instanceRegistry = InstanceRegistry(_newInstanceRegistryProxyAddress);
        emit InstanceRegistryUpdated(_newInstanceRegistryProxyAddress);
    }

    function setRelayerRegistry(address _newRelayerRegistryProxyAddress) external onlyGovernance {
        relayerRegistry = RelayerRegistry(_newRelayerRegistryProxyAddress);
        emit RelayerRegistryUpdated(_newRelayerRegistryProxyAddress);
    }

    /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ GETTERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

    function version() public pure virtual returns (string memory) {
        return "v2-oracle-manager";
    }
}
